import pandas as pd


def number_of_members(table: pd.DataFrame) -> pd.DataFrame:
    return (
        table.assign(Adult=lambda df: df["Age"] >= 14)
        .groupby(["Year", "ID"], as_index=False)
        .agg(
            Members=pd.NamedAgg("Member_Number", "count"),
            Adults=pd.NamedAgg("Adult", "sum"),
        )
        .assign(Childs=lambda df: df["Members"] - df["Adults"])
    )


def equivalence_scale(table: pd.DataFrame) -> pd.DataFrame:
    return table.assign(
        Household=1,
        Per_Capita=table["Members"],
        OECD=table["Adults"].multiply(0.7).add(0.3).add(table["Childs"].multiply(0.5)),
        OECD_Modified=table["Adults"]
        .multiply(0.5)
        .add(0.5)
        .add(table["Childs"].multiply(0.3)),
        Square_Root=table.eval("sqrt(Members)"),
    )


def create_season(table: pd.DataFrame) -> pd.DataFrame:
    seasons = {1: "Spring", 2: "Summer", 3: "Autumn", 4: "Winter"}
    season_series = pd.Series(
        data=pd.Categorical(
            table["Month"].floordiv(3).add(1).astype("Int16").map(seasons)
        ),
        index=table.index,
    )
    table["Season"] = season_series
    return table



"""HBSIR - Iran Household Budget Survey Data API.

This module provides an API for loading, processing, and analyzing
Iran's Household Budget Survey (HBSIR) data.

The key functions provided are:

- load_table: Load HBSIR data for a given table name and year range.

- add_classification: Add commodity/occupation classification codes.

- add_attribute: Add household attributes like urban/rural status. 

- add_weight: Add sampling weights.

- add_cpi: Join CPI index data.

- select: Filter table by urban/rural, province etc. 

- adjust_by_cpi: Adjust monetary values for inflation using CPI.

- adjust_by_equivalence_scale: Adjust by household composition.

- create_table_with_schema: Create table from schema.

- setup: Download, extract and process survey data.

- setup_config: Copy default config files.

- setup_metadata: Copy default metadata.

See the API docstrings for more details.

"""
# pylint: disable=too-many-arguments
# pylint: disable=unused-argument
# pylint: disable=too-many-locals

from typing import Iterable, Literal, overload
import shutil

import pandas as pd

from .core import archive_handler, data_cleaner, data_engine, decoder, metadata_reader
from . import external_data, utils
from .core.metadata_reader import (
    metadata,
    defaults,
    original_tables,
    _Attribute,
    _OriginalTable,
    _StandardTable,
    _Province,
    _Table,
    _Years,
    _EquivalenceScale,
)


def _extract_parameters(local_variables: dict) -> dict:
    return {key: value for key, value in local_variables.items() if value is not None}


@overload
def load_table(
    table_name: _OriginalTable,
    years: _Years = "last",
    form: Literal["processed", "cleaned", "raw"] | None = None,
    *,
    on_missing: Literal["error", "download", "create"] | None = None,
    redownload: bool | None = None,
    save_downloaded: bool | None = None,
    recreate: bool | None = None,
    save_created: bool | None = None,
) -> pd.DataFrame:
    ...


@overload
def load_table(
    table_name: _StandardTable,
    years: _Years = "last",
    form: Literal["processed"] | None = None,
    *,
    on_missing: Literal["error", "download", "create"] | None = None,
    redownload: bool | None = None,
    save_downloaded: bool | None = None,
    recreate: bool | None = None,
    save_created: bool | None = None,
) -> pd.DataFrame:
    ...


def load_table(
    table_name: _Table,
    years: _Years = "last",
    form: Literal["processed", "cleaned", "raw"] | None = None,
    *,
    on_missing: Literal["error", "download", "create"] | None = None,
    redownload: bool | None = None,
    save_downloaded: bool | None = None,
    recreate: bool | None = None,
    save_created: bool | None = None,
) -> pd.DataFrame:
    """Load a table for the given table name and year(s).

    This function loads original and standard tables.
    Original tables are survey tables and available in three types:
    original, cleaned and processed.

    - The 'raw' dataset contains the raw data, identical to the
    survey data, without any modifications.
    - The 'cleaned' dataset contains the raw data with added column
    labels, data types, and removal of irrelevant values, but no
    changes to actual data values.
    - The 'processed' dataset applies operations like adding columns,
    calculating durations, and standardizing tables across years.

    Standard tables are defined in this package to facilitate
    working with the data and are only available in processed form.

    For more information about available tables check the
    [tables wiki page](https://github.com/Iran-Open-Data/HBSIR/wiki/Tables).

    Parameters
    ----------
    table_name : str
        Name of the table to load.
    years : _Years, default "last"
        Year or list of years to load data for.
    dataset : str, default "processed"
        Which dataset to load from - 'processed', 'cleaned', or 'original'.
    on_missing : str, default "download"
        Action if data is missing - 'error', 'download', or 'create'
    recreate : bool, default False
        Whether to recreate the data instead of loading it
    redownload : bool, default False
        Whether to re-download the data instead of loading it
    save_downloaded : bool, default True
        Whether to save downloaded data
    save_created : bool, default True
        Whether to save newly created data

    Returns
    -------
    DataFrame
        Loaded data for the specified table and years.

    Examples
    --------
    >>> import hbsir
    >>> df = hbsir.load_table('food')
    # Loads processed 'food' table from original survey tables for
        latest available year
    >>> df = hbsir.load_table('Expenditures', '1399-1401')
    # Loads standard 'Expenditures' table for years 1399 - 1401

    Raises
    ------
    FileNotFoundError
        If data is missing and on_missing='error'.

    """
    metadata.reload_file("schema")
    parameters = _extract_parameters(locals())
    settings = data_engine.LoadTableSettings(**parameters)
    if settings.form == "raw":
        if table_name not in original_tables:
            raise ValueError
        years = utils.parse_years(years)
        table_parts = []
        for year in years:
            table_parts.append(data_cleaner.load_raw_data(table_name, year))
        table = pd.concat(table_parts)
    elif settings.form == "cleaned":
        if table_name not in original_tables:
            raise ValueError
        years = utils.parse_years(years)
        table_parts = []
        for year in years:
            table_parts.append(
                data_engine.TableHandler([table_name], year, settings)[table_name]
            )
        table = pd.concat(table_parts)
    else:
        table = data_engine.create_table(
            table_name=table_name,
            years=years,
            settings=settings,
        )
    return table


def create_table_with_schema(
    schema: str | dict,
    years: _Years = "last",
    *,
    on_missing: Literal["error", "download", "create"] | None = None,
    redownload: bool | None = None,
    save_downloaded: bool | None = None,
    recreate: bool | None = None,
    save_created: bool | None = None,
) -> pd.DataFrame:
    """Create and load table based on input schema.

    This function can be used in two ways:

    1) If a schema dictionary is passed, it will generate a DataFrame
    by loading or creating data that matches the schema.

    2) If a string table name is passed, it will load the table with
    that name from the external schema.

    Parameters
    ----------
    schema : dict or str
        Dictionary defining schema or string table name.
    years : str, optional
        Year(s) to load.
    on_missing : {'error', 'download', 'create'}, optional
        Action if data is missing.
    save_downloaded : bool, optional
        Whether to save downloaded data.
    save_created : bool, optional
       Whether to save newly created data.

    Returns
    -------
    DataFrame
        Loaded or created DataFrame.

    Examples
    --------
    >>> schema = {'table_list': ['food', 'tobacco'], 'years':'1398-1400'}
    >>> df = create_table_with_schema(schema)

    >>> table_name = 'User_Defined_Table'
    >>> df = create_table_with_schema(table_name)

    """
    metadata.reload_file("schema")
    parameters = _extract_parameters(locals())
    settings = data_engine.LoadTableSettings(**parameters)
    if isinstance(schema, str):
        return data_engine.create_table(schema, years, settings)

    if years is None and "years" in schema:
        years = schema["years"]
    if "table_list" in schema:
        table_name = "_Input_Table"
        metadata.schema[table_name] = schema
    elif all("table_list" in table_schema for table_schema in schema.values()):
        table_name = list(schema.keys())[-1]
        metadata.schema.update(schema)
    else:
        raise NameError
    return data_engine.create_table(table_name, years, settings)


def add_classification(
    table: pd.DataFrame,
    name: str = "original",
    classification_type: Literal["commodity", "industry", "occupation"] | None = None,
    *,
    aspects: Iterable[str] | None = None,
    levels: Iterable[int] | int | None = None,
    column_names: Iterable[str] | str | None = None,
    drop_value: bool | None = None,
    missing_value_replacements: dict[str, str] | None = None,
    code_col: str | None = None,
    year_col: str | None = None,
) -> pd.DataFrame:
    """Add classification columns to table.

    Classifies codes in the table using specified classification system.

    Supported systems:

    - 'commodity': Classifies commodity codes
    - 'industry': Classifies industry codes
    - 'occupation': Classifies occupation codes

    Parameters
    ----------
    table : DataFrame
        Table containing code column to classify.
    name : str, default 'original'
        Name of classification to apply.
    classification_type : str, optional
        Type of classification system.
    aspects : list, optional
        Aspects of classification to add as columns.
    levels : list of int, optional
        Number of digits for each classification level.
    column_names: list of str, optional
        Names of output columns.
    drop_value : bool, optional
        Whether to drop unclassified values.
    missing_value_replacements : dict, optional
        Replacements for missing values in columns.
    code_col: str, optional
        Name of the code column.
    year_col: str, optional
        Name of the year column.

    Returns
    -------
    DataFrame
        Table with added classification columns.

    """
    parameters = _extract_parameters(locals())
    if "classification_type" not in parameters:
        if "code_column_name" in parameters:
            if table[parameters["code_column_name"]].le(10_000).mean() < 0.9:
                class_type = "occupation"
            else:
                class_type = "commodity"
        elif metadata_reader.defaults.columns.commodity_code in table.columns:
            class_type = "commodity"
        elif metadata_reader.defaults.columns.job_code in table.columns:
            class_type = "occupation"
        else:
            raise ValueError("Missing Code Column")
        parameters["classification_type"] = class_type
    settings = decoder.DecoderSettings(**parameters)
    table = decoder.Decoder(table=table, settings=settings).add_classification()
    return table


def add_attribute(
    table: pd.DataFrame,
    name: _Attribute,
    *,
    aspects: Iterable[str] | str | None = None,
    column_names: Iterable[str] | str | None = None,
    id_col: str | None = None,
    year_col: str | None = None,
) -> pd.DataFrame:
    """Add household attributes to table based on ID.

    Takes a DataFrame with a household ID column and adds columns
    for the specified attribute such as urban/rural status or province.

    Supported attributes:

    - 'Urban_Rural': Urban or rural classification
    - 'Province': Province name
    - 'County': County name

    Parameters
    ----------
    table : DataFrame
        DataFrame containing ID column.
    name : str
        Name of attribute to add.
    aspects: list of str, optional
        Aspects of attribute to add as columns.
    column_names: list of str, optional
        Output column names.
    id_col: str, optional
        Name of ID column.
    year_col: str, optional
        Name of year column.

    Returns
    -------
    DataFrame
        Input DataFrame with added attribute columns.

    """
    parameters = _extract_parameters(locals())
    settings = decoder.IDDecoderSettings(**parameters)
    table = decoder.IDDecoder(table=table, settings=settings).add_attribute()
    return table


def select(
    table: pd.DataFrame,
    *,
    urban_rural: Literal["Urban", "Rural"] | None = None,
    province: Iterable[_Province] | _Province | None = None,
    county: Iterable[str] | str | None = None,
) -> pd.DataFrame:
    """Select subset of table based on criteria.

    Filters table by urban/rural, province, or county.

    Decodes IDs to attributes if needed and removes them after.

    Parameters
    ----------
    table : pd.DataFrame
        Input DataFrame to filter.

    urban_rural : Literal["Urban", "Rural"], optional
        Keep only urban or rural households.

    province : Province, optional
        Keep only given province(s).

    county : str or list, optional
        Keep only given county/counties.

    Returns
    -------
    pd.DataFrame
        Subset of input table meeting criteria.

    """
    if urban_rural is not None:
        table = (
            table.pipe(add_attribute, "Urban_Rural")
            .query(f"Urban_Rural == '{urban_rural}'")
            .drop(columns="Urban_Rural")
        )
    if county is not None:
        if isinstance(county, str):
            county = [county]
        table = (
            table.pipe(add_attribute, "County")
            .query(f"County in {county}")
            .drop(columns="County")
        )
    elif province is not None:
        if isinstance(province, str):
            province = [province]
        table = (
            table.pipe(add_attribute, "Province")
            .query(f"Province in [{province}]")
            .drop(columns="Province")
        )
    return table


def add_weight(
    table: pd.DataFrame, adjust_for_household_size: bool = False
) -> pd.DataFrame:
    """Add sampling weights to the table.

    Loads appropriate sample weights for each year in the table and merges
    them onto the table. Sample weights can optionally be adjusted
    for household size.

    Weights for years prior to 1395 are loaded from external parquet data,
    while weights for 1395 onward come from the household_information table.

    Parameters
    ----------
    table : pd.DataFrame
        Input data table, containing a column of year values
    adjust_for_household_size : bool, default False
        Whether to adjust weights by household size
    year_column_name : str, default "Year"
        Name of column in `table` that contains the year

    Returns
    -------
    table : pd.DataFrame
        Input `table` with 'Weight' column added

    """
    table = data_engine.add_weights(table, adjust_for_household_size)
    return table


def add_cpi(
    table: pd.DataFrame,
    *,
    data_source: Literal["SCI"] = "SCI",
    base_year: Literal[1400] = 1400,
    frequency: Literal["Annual"] = "Annual",
    separate_by: Literal["Urban_Rural"] = "Urban_Rural",
) -> pd.DataFrame:
    """Add CPI values to the table

    Parameters
    ----------
    table : pd.DataFrame
        Input DataFrame to add CPI values to.
    data_source : str, default "SCI"
        Source of CPI data, either "SCI" or "CBS".
    base_year : int, default 1400
        Base year for CPI.
    frequency : str, default "Annual"
        Frequency of CPI data, either "Annual" or "Monthly".
    separate_by : str, default "Urban_Rural"
        Column to separate CPI by, either "Urban_Rural" or None.

    Returns
    -------
    DataFrame
        Input `table` with 'CPI' column added.

    """
    cpi = external_data.load_table(
        f"CPI_{base_year}",
        data_source=data_source,
        frequency=frequency,
        separate_by=separate_by,
        reset_index=False,
    )

    if separate_by == "Urban_Rural":
        urban_rural_available = True
        if "Urban_Rural" not in table.columns:
            table = add_attribute(table, "Urban_Rural")
            urban_rural_available = False

        table = table.join(cpi, on=["Urban_Rural", "Year"])

        if not urban_rural_available:
            table = table.drop(columns="Urban_Rural")
    else:
        table = table.join(cpi, on=["Year"])

    return table


def adjust_by_cpi(
    table: pd.DataFrame, columns: list[str] | None = None, **kwargs
) -> pd.DataFrame:
    """Adjust columns in the table by the CPI

    Divides the specified columns of the DataFrame by the CPI column.
    If no columns specified, divides common monetary value columns
    like 'Expenditure', 'Income' etc.

    Parameters
    ----------
    table : DataFrame
        DataFrame containing a "CPI" column.
    columns : List[str], optional
        List of columns names to divide by CPI.
    **kwargs
        Additional keyword arguments passed to add_cpi.

    Returns
    -------
    DataFrame
        Input `table` with `columns` divided by 'CPI'.

    """
    default_columns = metadata_reader.defaults.columns.nominals
    if columns is None:
        columns = [column for column in default_columns if column in table.columns]

    cpi_available = True
    if "CPI" not in table.columns:
        table = add_cpi(table, **kwargs)
        cpi_available = False
    table.loc[:, columns] = (
        table.loc[:, columns]
        .divide(table["CPI"], axis="index")
        .multiply(100, axis="index")
    )
    if not cpi_available:
        table = table.drop(columns="CPI")

    return table


def adjust_by_equivalence_scale(
    table: pd.DataFrame,
    columns: list[str],
    equivalence_scale: _EquivalenceScale = "Per_Capita",
) -> pd.DataFrame:
    """Adjust columns by household equivalence scale.

    Divides specified columns by household equivalence scale
    to account for household size and composition.

    Supported scales are:
        - Household: No adjustment (household total)
        - Per_Capita: Divide by household size
        - OECD: OECD equivalence scale
        - OECD_Modified: Modified OECD scale
        - Square_Root: Square root of household size

    Loads equivalence scales from Equivalence_Scale table.

    Parameters
    ----------
    table : DataFrame
        Input DataFrame with household ID.
    columns : List[str]
        Columns to adjust.
    equivalence_scale : _EquivalenceScale, default "Per_Capita"
        Equivalence scale to use.

    Returns
    -------
    DataFrame
        Table with adjusted columns.

    """
    years = table[defaults.columns.year].unique().tolist()

    equivalence_table = load_table("Equivalence_Scale", years=years).set_index(
        [defaults.columns.year, defaults.columns.household_id]
    )

    table = table.join(
        equivalence_table[equivalence_scale], on=["Year", "ID"], how="left"
    )
    table[columns] = table[columns].div(table[equivalence_scale], axis="index")
    table = table.drop(columns=equivalence_scale)

    return table


def setup(
    years: _Years = "last",
    method: Literal["create", "download"] = "create",
    table_names: _OriginalTable | Iterable[_OriginalTable] | Literal["all"] = "all",
    replace: bool = False,
) -> None:
    """Download, extract, and process HBSIR data.

    Handles downloading and extracting the HBSIR data files,
    saving them locally, and processing tables.

    This sets up the raw data for further analysis. It:

    - Downloads and extracts archive files using archive_handler.
    - Cleans tables and saves as Parquet using data_cleaner.

    Parameters
    ----------
    years : str/list/int, default "last"
        Year(s) to download and process.
    method : str, default "create"
        "create" to process raw tables or
        "download" to download processed tables.
    table_names : str/list, default "all"
        Tables to process. "all" for all tables.
    replace : bool, default False
        Whether to overwrite existing files.

    Examples
    --------
    setup(1399)
        Process data for 1399.

    setup([1390,1400], table_names=['food'])
        Process food table for 1390 and 1400.

    setup('1380-1390', method='download')
        Download processed tables for 1380-1390.

    """
    if method == "create":
        archive_handler.setup(years, replace)
        data_cleaner.save_cleaned_tables(table_names, years)
    else:
        utils.download_processed_data(table_names, years)


def setup_config(replace=False) -> None:
    """Copy default config file to data directory.

    Copies the default config file 'settings-sample.yaml' from
    the package directory to 'settings.yaml' in the root data
    directory.

    Overwrites any existing file if replace=True.

    Parameters
    ----------
    replace : bool, default False
        Whether to overwrite existing config file.

    """
    src = defaults.package_dir.joinpath("config", "settings-sample.yaml")
    dst = defaults.root_dir.joinpath("settings.yaml")
    if (not dst.exists()) or replace:
        shutil.copy(src, dst)


def setup_metadata(replace=False) -> None:
    """Copy default metadata files to data directory.

    Copies metadata files like schema.csv and info.csv from the
    package metadata folder to the root data/metadata folder.

    Overwrites any existing files if replace=True.

    Parameters
    ----------
    replace : bool, default False
        Whether to overwrite existing metadata files.

    """
    src_folder = defaults.package_dir.joinpath("metadata")
    dst_folder = defaults.root_dir.joinpath("metadata")
    if not dst_folder.exists():
        dst_folder.mkdir()
    for src in src_folder.iterdir():
        dst = dst_folder.joinpath(src.name)
        if (not dst.exists()) or replace:
            shutil.copy(src, dst)



"""
Module for cleaning raw data into proper format
"""

from pathlib import Path
from typing import Iterable, Literal

from tqdm import tqdm
import pandas as pd
import numpy as np

from .. import utils

from .metadata_reader import (
    defaults,
    metadata,
    original_tables,
    _OriginalTable,
    _Years,
)


def load_raw_data(
    table_name: _OriginalTable, year: int, urban: bool | None = None
) -> pd.DataFrame:
    """Reads CSV file(s) and returns DataFrame for table and year.

    Reads the CSV file(s) containing the specified table data for the
    given year and urban/rural category.

    Concatenates and returns the contents as a pandas DataFrame.

    Parameters
    ----------
    table_name : str
        Name of the table to read.

    year : int
        Year of the data to read.

    urban : bool, optional
        Whether to read only urban, rural or both data.
        If None, reads both urban and rural.

    Returns
    -------
    DataFrame
        Concatenated table data from the CSV file(s).

    Raises
    ------
    FileNotFoundError
        If CSV file(s) not found.

    ValueError
        If invalid table name, year, or corrupt metadata.

    Examples
    --------
    >>> df = load_raw_data('food', 1393)
    >>> df = load_raw_data('population', 1390, urban=True)

    """
    urban_stats = [True, False] if urban is None else [urban]
    tables = []
    for is_urban in urban_stats:
        file_path = _build_file_path(table_name, year, is_urban)
        tables.append(pd.read_csv(file_path, low_memory=False))
    table = pd.concat(tables, ignore_index=True)
    return table


def _build_file_path(table_name: str, year: int, is_urban: bool) -> Path:
    urban_rural = "U" if is_urban else "R"
    year_string = year % 100 if year < 1400 else year
    table_metadata = _get_table_metadata(table_name, year, is_urban)
    file_code = utils.resolve_metadata(table_metadata["file_code"], year)
    if file_code is None:
        raise ValueError(f"Table {table_name} is not available for year {year}")
    file_name = f"{urban_rural}{year_string}{file_code}.csv"
    file_path = defaults.extracted_data.joinpath(str(year), file_name)
    return file_path


# pylint: disable=unsubscriptable-object
# pylint: disable=unsupported-membership-test
def _get_table_metadata(
    table_name: str, year: int, is_urban: bool | None = None
) -> dict:
    table_metadata = metadata.tables[table_name]
    table_metadata = utils.resolve_metadata(table_metadata, year)
    assert isinstance(table_metadata, dict)

    if is_urban is True:
        if "urban" in table_metadata:
            table_metadata = table_metadata["urban"]
    if is_urban is False:
        if "rural" in table_metadata:
            table_metadata = table_metadata["rural"]

    table_metadata["table_name"] = table_name
    table_metadata["year"] = year
    table_metadata["is_urban"] = is_urban

    return table_metadata


def open_and_clean_table(table_name: _OriginalTable, year: int) -> pd.DataFrame:
    """Cleans table data using metadata transformations.

    Loads raw table data, applies cleaning ops based on metadata,
    and concatenates urban and rural tables.

    Useful as a preprocessing step before further analysis.
    Called by save_processed_tables() to clean each table.

    Parameters
    ----------
    table_name : _OriginalTable
        Name of table to clean.

    year : int
        Year of data to clean.

    Returns
    -------
    DataFrame
        Cleaned concatenated table data.

    """
    cleaned_table_list = []
    for is_urban in [True, False]:
        table = load_raw_data(table_name, year, is_urban)
        table_metadata = _get_table_metadata(table_name, year, is_urban)
        cleaned_table = _apply_metadata_to_table(table, table_metadata)
        cleaned_table_list.append(cleaned_table)
    final_table = pd.concat(cleaned_table_list, ignore_index=True)
    return final_table


def _apply_metadata_to_table(table: pd.DataFrame, table_metadata: dict) -> pd.DataFrame:
    cleaned_table = pd.DataFrame()
    for column_name, column in table.items():
        assert isinstance(column_name, str)
        column_metadata = _get_column_metadata(table_metadata, column_name.upper())
        if column_metadata == "drop":
            continue
        if column_metadata == "error":
            raise ValueError(
                f"Error: The column '{column_name}' was not found in the metadata."
            )
        column = _apply_metadata_to_column(column, column_metadata)
        cleaned_table[column_metadata["new_name"]] = column
    return cleaned_table


def _get_column_metadata(
    table_metadata: dict, column_name: str
) -> dict | Literal["drop", "error"]:
    table_settings = _get_table_settings(table_metadata)
    year = table_metadata["year"]
    columns_metadata = table_metadata["columns"]
    columns_metadata = utils.MetadataVersionResolver(
        columns_metadata, year
    ).get_version()
    if not isinstance(columns_metadata, dict):
        raise ValueError(
            f"Unvalid metadata for column {column_name}: \n {columns_metadata}"
        )
    if column_name in columns_metadata:
        column_metadata = columns_metadata[column_name]
        if not (isinstance(column_metadata, dict) or column_metadata == "drop"):
            print(table_metadata)
            raise ValueError(f"Metadata for column {column_name} is not valid")
    else:
        column_metadata: Literal["drop", "error"] = table_settings["missings"]
        if column_metadata not in ["drop", "error"]:
            raise ValueError("Missing treatment is not valid")
    return column_metadata


def _get_table_settings(table_metadata: dict) -> dict:
    default_table_settings = metadata.tables["default_table_settings"]
    try:
        table_settings = table_metadata["settings"]
    except KeyError:
        return default_table_settings
    for key in default_table_settings:
        if key not in table_settings:
            table_settings[key] = default_table_settings[key]
    return table_settings


def _apply_metadata_to_column(column: pd.Series, column_metadata: dict) -> pd.Series:
    if ("replace" in column_metadata) and (column_metadata["replace"] is not None):
        column = column.replace(column_metadata["replace"])
    column = _apply_type_to_column(column, column_metadata)
    return column


def _apply_type_to_column(column: pd.Series, column_metadata: dict) -> pd.Series:
    column = _general_cleaning(column)
    new_column = pd.Series(np.nan, index=column.index)
    if ("type" not in column_metadata) or (column_metadata["type"] == "string"):
        new_column = column.copy()
    elif column_metadata["type"] == "boolean":
        new_column = column.astype("Int32") == column_metadata["true_condition"]
    elif column_metadata["type"] in ("unsigned", "integer", "float"):
        new_column = pd.to_numeric(column, downcast=column_metadata["type"])
    elif column_metadata["type"] == "category":
        new_column = column.astype("Int32").astype("category")
        new_column = new_column.cat.rename_categories(column_metadata["categories"])
    return new_column


def _general_cleaning(column: pd.Series):
    if pd.api.types.is_numeric_dtype(column):
        return column
    chars_to_remove = r"\n\r\,\@\+\*\[\]\_\?"
    try:
        column = column.str.replace(chr(183), ".").str.rstrip(".")
        column = column.str.replace(f"[{chars_to_remove}]+", "", regex=True)
        column = column.str.replace(r"\b\-", "", regex=True)
        column = column.replace(r"^[\s\.\-]*$", np.nan, regex=True)
    except AttributeError:
        pass
    return column


def save_cleaned_tables(
    table_names: _OriginalTable | Iterable[_OriginalTable] | Literal["all"] = "all",
    years: _Years = "all",
) -> None:
    """Saves cleaned table data to Parquet files.

    Cleans, processes and saves the specified tables for the given years
    as Parquet files in the processed_data directory.

    Parameters
    ----------
    table_names : _OriginalTable or Iterable[_OriginalTable], optional
        Names of tables to process.
        Default is to process all tables.

    years : _Years, optional
        Years of data to process.
        Default is "all" years.

    """
    table_names = original_tables if table_names == "all" else table_names
    table_year = utils.construct_table_year_pairs(table_names, years)
    pbar = tqdm(total=len(table_year), desc="Preparing ...", unit="Table")
    for _table_name, year in table_year:
        pbar.update()
        pbar.desc = f"Table: {_table_name}, Year: {year}"
        table = open_and_clean_table(_table_name, year)
        Path(defaults.processed_data).mkdir(exist_ok=True)
        table.to_parquet(
            defaults.processed_data.joinpath(f"{year}_{_table_name}.parquet"),
            index=False,
        )
    pbar.close()



import pandas as pd

from ..core.metadata_reader import defaults, _Groupby
from .. import api
from .quantile import add_decile, add_percentile


def weighted_average(
    table: pd.DataFrame, columns: list[str] | None = None, weight_col: str | None = None
) -> pd.Series:
    """Calculate the weighted average of columns in a DataFrame.

    Parameters
    ----------
    table : DataFrame
        Input DataFrame containing the columns to calculate
        weighted average for and the weight column.
    weight_col : str, default 'Weight'
        The name of the column containing the weights.

    Returns
    -------
    pandas Series
        A Series containing the weighted average of each column.

    Raises
    ------
    ValueError
        If the weight column is not in the table.

    Examples
    --------
    >>> df = pd.DataFrame({
            'Col1': [1, 2, 3],
            'Col2': [25, 15, 10],
            'Weight': [0.25, 0.25, 0.5]})
    >>> weighted_average(df)
        Col1     2.25
        Col2    15.00
        dtype: float64

    """
    weight_col = defaults.columns.weight if weight_col is None else weight_col
    if weight_col not in table.columns:
        raise ValueError(f"Weight column {weight_col} not in table")
    if columns is None:
        columns = [
            col
            for col in table.select_dtypes("number").columns
            if col in table.columns
            if col not in defaults.columns.groupby
            if col not in [defaults.columns.household_id, weight_col, "index"]
        ]
    table[columns] = table[columns].multiply(table[weight_col], axis="index")
    columns_summation = table.sum()
    results = (
        columns_summation.loc[columns]
        .divide(columns_summation[weight_col])
        .loc[columns]
    )
    return results


def average_table(
    table: pd.DataFrame,
    columns: list[str] | None = None,
    groupby: list[_Groupby] | _Groupby | None = None,
    weight_col: str | None = None,
    weighted: bool = True,
) -> pd.DataFrame:
    if isinstance(table.columns, pd.MultiIndex):
        is_multi_index = True
        column_names = table.columns.names
        table.columns = table.columns.to_flat_index()
    else:
        is_multi_index = False
        column_names = None

    table = table.reset_index().copy()

    if groupby is None:
        groupby = [col for col in table.columns if col in defaults.columns.groupby]
    elif isinstance(groupby, str):
        groupby = [groupby]
    if groupby is None:
        groupby = []

    for groupby_column in ("Urban_Rural", "Province", "County"):
        if groupby_column not in table.columns:
            table = api.add_attribute(table, groupby_column)
    if "Percentile" in groupby:
        table = add_percentile(table)
    elif "Decile" in groupby:
        table = add_decile(table)

    weight_col = defaults.columns.weight if weight_col is None else weight_col

    if not weighted:
        table["weight_col"] = 1
    elif weight_col not in table.columns:
        table = api.add_weight(table)

    if len(groupby) == 0:
        row = weighted_average(table, columns=columns, weight_col=weight_col)
        result = pd.DataFrame([row])
    else:
        result = table.groupby(groupby).apply(weighted_average, columns, weight_col)

    if is_multi_index:
        result.columns = pd.MultiIndex.from_tuples(result.columns, names=column_names)

    return result



"""Main API for Household Budget Survey of Iran (HBSIR) data.

This module exposes the main functions for loading, processing, and  
analyzing the household survey data.

The key functions are:

- load_table: Load data for a given table name and year range
- add_classification: Add COICOP commodity classification codes  
- add_attribute: Add household attributes like urban/rural
- add_weight: Add sampling weights

The load_table function handles fetching data from different sources
like original CSVs or preprocessed Parquet files. It provides options 
for configuring behavior when data is missing.

Sample usage:

    import hbsir
    
    df = hbsir.load_table('food', [1399, 1400])
    df = hbsir.add_classification(df, 'original')
    df = hbsir.add_attribute(df, 'Urban_Rural')
    
See API documentation for more details.
"""


from . import (
    calculator,
    external_data,
    hbsframe,
    utils,
)
from .api import (
    load_table,
    create_table_with_schema,
    add_classification,
    add_attribute,
    select,
    add_weight,
    add_cpi,
    adjust_by_cpi,
    adjust_by_equivalence_scale,
    setup,
    setup_config,
    setup_metadata,
)

__version__ = "0.3.0"

__all__ = [
    "load_table",
    "create_table_with_schema",
    "add_attribute",
    "select",
    "add_classification",
    "add_weight",
    "add_cpi",
    "adjust_by_cpi",
    "adjust_by_equivalence_scale",
    "setup",
    "setup_config",
    "setup_metadata",
    "calculator",
    "external_data",
    "hbsframe",
    "utils",
]



"""Main module for loading and transforming HBSIR data.

This module provides the central interfaces for loading raw Iranian 
household data tables, transforming them, and constructing cleaned
derivative tables for analysis.

Key functions:

- extract_dependencies - Get dependencies for building a table 
- TableHandler - Loads multiple dependency tables
- Pipeline - Applies a sequence of transform steps to a table
- TableFactory - Loads and builds tables from different sources
- create_table - Constructs a table by loading multiple years  
- load_weights - Loads sample weights for a given year
- add_weights - Adds weights to a table

The module focuses on ETL (Extract, Transform, Load) functions to go 
from raw provided data tables to cleaned analytic tables.

Relies on metadata schema and configuration for how to process tables.

"""

import re
from pathlib import Path
from typing import Literal, Iterable
from types import ModuleType
import importlib
from concurrent.futures import ThreadPoolExecutor

import pandas as pd
from pydantic import BaseModel
import yaml

from . import decoder, metadata_reader

from .. import utils, external_data
from .metadata_reader import (
    defaults,
    metadata,
    original_tables,
    _Years,
    _OriginalTable,
)
from .data_cleaner import open_and_clean_table


def extract_dependencies(table_name: str, year: int) -> dict:
    """Extract the dependencies of a table based on the metadata schema.

    For the given table name and year, traverses the schema metadata to find
    all upstream dependencies that are required to construct the table.

    Recursively extracts dependencies of dependencies until only base tables
    remain. Base tables have their file size stored instead of further dependencies.

    Parameters
    ----------
    table_name : str
        Name of the target table to extract dependencies for

    year : int
        Year to extract schema dependencies for

    Returns
    -------
    dependencies : dict
        Dictionary with dependencies in the format:
        {table_name: {"dependencies": {dep1: {}, dep2: {}}},
         table_name2: {"size": 1024}}
    """
    table_list = [table_name]
    dependencies: dict[str, dict] = {}
    while len(table_list) > 0:
        table = table_list.pop(0)
        if "table_list" in metadata.schema[table]:
            dependencies[table] = metadata.schema[table]
            upstream_tables = utils.MetadataVersionResolver(
                metadata.schema[table]["table_list"], year=year
            ).get_version()
            if isinstance(upstream_tables, str):
                upstream_tables = [upstream_tables]
            assert isinstance(upstream_tables, list)
            table_list.extend(upstream_tables)
        elif table in original_tables:
            file_name = f"{year}_{table}.parquet"
            local_path = defaults.processed_data.joinpath(file_name)
            size = local_path.stat().st_size if local_path.exists() else None
            dependencies[table] = {"size": size}
        else:
            raise ValueError
    return dependencies


class LoadTableSettings(BaseModel):
    form: Literal["processed", "cleaned", "raw"] = metadata_reader.settings[
        ("functions_defaults", "load_table", "dataset")
    ]
    on_missing: Literal["error", "download", "create"] = metadata_reader.settings[
        ("functions_defaults", "load_table", "on_missing")
    ]
    save_downloaded: bool = metadata_reader.settings[
        ("functions_defaults", "load_table", "save_downloaded")
    ]
    redownload: bool = metadata_reader.settings[
        ("functions_defaults", "load_table", "redownload")
    ]
    save_created: bool = metadata_reader.settings[
        ("functions_defaults", "load_table", "save_created")
    ]
    recreate: bool = metadata_reader.settings[
        ("functions_defaults", "load_table", "recreate")
    ]


class TableHandler:
    """Handles loading multiple tables from parquet.

    Loads a set of tables by reading from local parquet files or downloading
    or generating if missing. Provides access to the tables via indexing.

    Attributes
    ----------
    table_list : list of str
        List of table names to load
    year : int
        Year for tables
    settings : LoadTable
        Settings for how to load the tables
    tables : dict of DataFrames
        Loaded tables keyed by table name

    """

    def __init__(
        self,
        table_list: Iterable[_OriginalTable],
        year: int,
        settings: LoadTableSettings | None = None,
    ) -> None:
        self.table_list = table_list
        self.year = year
        self.settings = settings if settings is not None else LoadTableSettings()
        self.tables: dict[str, pd.DataFrame] = self.setup()

    def __getitem__(self, table_name: _OriginalTable) -> pd.DataFrame:
        """Get a table by name.

        Parameters
        ----------
        table_name : str
            Name of table to retrieve

        Returns
        -------
        table : DataFrame
            Table loaded for the given name

        """
        return self.tables[table_name]

    def get(
        self, names: _OriginalTable | Iterable[_OriginalTable]
    ) -> list[pd.DataFrame]:
        """Get multiple tables by name.

        Parameters
        ----------
        names : str or list of str
            Name(s) of tables to retrieve

        Returns
        -------
        tables : list of DataFrames
            Requested tables loaded

        """
        names = [names] if isinstance(names, str) else names
        return [self[name] for name in names]

    def setup(self) -> dict[str, pd.DataFrame]:
        """Set up the handler by loading all tables.

        Loads all of the configured tables in parallel using a
        ThreadPoolExecutor.

        Returns
        -------
        tables : dict of DataFrames
            Dictionary of the loaded tables by name.

        """
        with ThreadPoolExecutor(max_workers=6) as executer:
            tables = zip(
                self.table_list, executer.map(self.read_table, self.table_list)
            )
        return dict(tables)

    def read_table(self, table_name: _OriginalTable) -> pd.DataFrame:
        """Read a single table by name.

        Parameters
        ----------
        table_name : str
            Name of the table to load

        Returns
        -------
        table : DataFrame
            Loaded table data

        """

        if self.settings.recreate:
            table = self._create_table(table_name)
        elif self.settings.redownload:
            table = self._download_table(table_name)
        elif self.get_local_path(table_name).exists():
            table = self._load_table(table_name)
        elif self.settings.on_missing == "create":
            table = self._create_table(table_name)
        elif self.settings.on_missing == "download":
            table = self._download_table(table_name)
        else:
            raise FileNotFoundError

        table.attrs["table_name"] = table_name
        table.attrs["year"] = self.year
        return table

    def get_local_path(self, table_name) -> Path:
        file_name = f"{self.year}_{table_name}.parquet"
        local_path = defaults.processed_data.joinpath(file_name)
        return local_path

    def _create_table(self, table_name: _OriginalTable) -> pd.DataFrame:
        table = open_and_clean_table(table_name, self.year)
        if self.settings.save_created:
            table.to_parquet(self.get_local_path(table_name))
        return table

    def _download_table(self, table_name: _OriginalTable) -> pd.DataFrame:
        table = pd.read_parquet(
            f"{defaults.online_dir}/parquet_files/{self.year}_{table_name}.parquet"
        )
        if self.settings.save_downloaded:
            table.to_parquet(self.get_local_path(table_name))
        return table

    def _load_table(self, table_name: _OriginalTable) -> pd.DataFrame:
        return pd.read_parquet(self.get_local_path(table_name))


class Pipeline:
    """Applies a sequence of transformation steps to a DataFrame.

    This class allows chaining together a set of predefined steps
    for cleaning, transforming, and processing a DataFrame representing
    a table of data. The steps are configured by passing a list of
    operations which are applied in sequence.

    Attributes
    ----------
    table : DataFrame
        The input DataFrame that steps are applied to
    steps : list
        The sequence of step functions to apply
    properties : dict
        Additional properties passed to steps

    """

    def __init__(
        self, table: pd.DataFrame, steps: list, table_name: str, year: int
    ) -> None:
        self.table = table
        self.table_name = table_name
        self.year = year
        self.steps = steps
        self.modules: dict[str, ModuleType] = {}

    def run(self) -> pd.DataFrame:
        """Run the pipeline on the table.

        Iterates through the step functions in the pipeline
        and applies them sequentially to transform the table.

        Returns
        -------
        table : DataFrame
            The transformed table after applying all steps.
        """
        for step in self.steps:
            if step is None:
                continue
            method_name, method_input = self._extract_method_name(step)
            if method_input is None:
                getattr(self, f"_{method_name}")()
            else:
                getattr(self, f"_{method_name}")(method_input)
        return self.table

    def _extract_method_name(self, instruction):
        if isinstance(instruction, str):
            method_name = instruction
            method_input = None
        elif isinstance(instruction, dict):
            method_name, method_input = list(instruction.items())[0]
        else:
            raise TypeError
        return method_name, method_input

    def _add_year(self) -> None:
        self.table["Year"] = self.year

    def _add_table_name(self) -> None:
        self.table["Table_Name"] = self.table_name

    def _add_weights(self) -> None:
        self.table = add_weights(self.table)

    def _add_classification(self, method_input: dict | None = None) -> None:
        if method_input is None:
            return
        settings = decoder.DecoderSettings(**method_input)
        self.table = decoder.Decoder(self.table, settings).add_classification()

    def _add_attribute(self, method_input: dict | None = None) -> None:
        if method_input is None:
            return
        settings = decoder.IDDecoderSettings(**method_input)
        self.table = decoder.IDDecoder(self.table, settings).add_attribute()

    def _apply_order(self, method_input: list):
        new_order = [
            column if isinstance(column, str) else list(column.keys())[0]
            for column in method_input
        ]
        types = {
            list(column.keys())[0]: list(column.values())[0]
            for column in method_input
            if isinstance(column, dict)
        }

        self.table = self.table[list(new_order)].astype(types)

    def _create_column(self, method_input: dict | None = None) -> None:
        if method_input is None:
            return
        column_name = method_input["name"]
        if method_input["type"] == "numerical":
            expression = method_input["expression"]
            self.__apply_numerical_instruction(column_name, expression)
        elif method_input["type"] == "categorical":
            categories = method_input["categories"]
            self.__apply_categorical_instruction(column_name, categories)

    def __apply_numerical_instruction(self, column_name, expression: int | str) -> None:
        if isinstance(expression, int):
            self.table.loc[:, column_name] = expression
            return
        columns_names = re.split(r"[\+\-\*\/\s\.\(\)]+", expression)
        columns_names = [
            name for name in columns_names if not (name.isnumeric() or (name is None))
        ]
        self.table[column_name] = (
            self.table[columns_names].fillna(0).eval(expression, engine="python")
        )

    def __apply_categorical_instruction(
        self, column_name: str, categories: dict
    ) -> None:
        if column_name in self.table.columns:
            categorical_column = self.table[column_name].copy()
        else:
            categorical_column = pd.Series(index=self.table.index, dtype="category")

        for category, condition in categories.items():
            filt = self.__construct_filter(column_name, condition)
            categorical_column = categorical_column.cat.add_categories([category])
            categorical_column.loc[filt] = category

        self.table[column_name] = categorical_column

    def __construct_filter(self, column_name, condition) -> pd.Series:
        if condition is None:
            filt = self.table.index.to_series()
        elif isinstance(condition, str):
            filt = self.table[column_name] == condition
        elif isinstance(condition, list):
            filt = self.table[column_name].isin(condition)
        elif isinstance(condition, dict):
            filts = []
            for other_column, value in condition.items():
                if isinstance(value, (bool, str)):
                    filts.append(self.table[other_column] == value)
                elif isinstance(value, list):
                    filts.append(self.table[other_column].isin(value))
                else:
                    raise KeyError
            filt_sum = pd.concat(filts, axis="columns").sum(axis="columns")
            filt = filt_sum == len(condition)
        else:
            raise KeyError
        return filt

    def _apply_filter(self, conditions: str | list[str] | None = None) -> None:
        if conditions is None:
            return
        conditions = [conditions] if isinstance(conditions, str) else conditions
        for condition in conditions:
            self.table = self.table.query(condition)

    def _apply_pandas_function(self, method_input: str | None = None) -> None:
        if method_input is None:
            return
        method_input = "self.table" + method_input.replace("\n", "")
        table = pd.eval(method_input, target=self.table, engine="python")
        assert isinstance(table, pd.DataFrame)
        self.table = table

    def _apply_external_function(self, method_input: str | None = None) -> None:
        if method_input is None:
            return
        module_name, func_name = method_input.rsplit(".", 1)
        self.__load_module(module_name)
        func = getattr(self.modules[module_name], func_name)
        self.table = func(self.table)

    def __load_module(self, module_name: str) -> None:
        if module_name not in self.modules:
            self.modules[module_name] = importlib.import_module(module_name)

    def _join(self, method_input: dict | str | None = None):
        if method_input is None:
            return
        if isinstance(method_input, str):
            table_name = method_input
            columns = ["Year", "ID"]
            years = None
        elif isinstance(method_input, dict):
            table_name = method_input["table_name"]
            columns = method_input["columns"]
            years = method_input.get("year", None)
        else:
            raise TypeError
        years = list(self.table["Year"].unique())
        other_table = create_table(table_name, years)
        self.table = self.table.merge(other_table, on=columns)


class TableFactory:
    """Builds DataFrames representing tables of data.

    This class handles loading or constructing DataFrames representing
    different tables of data. It builds tables either from original
    source parquet files, by querying cached results, or by dynamically
    constructing the table from other tables based on a schema.

    """

    def __init__(
        self,
        table_name: str,
        year: int,
        settings: LoadTableSettings | None = None,
    ):
        self.table_name = table_name
        self.year = year
        self.settings = settings if settings is not None else LoadTableSettings()
        schema = utils.MetadataVersionResolver(metadata.schema, year).get_version()

        if isinstance(schema, dict):
            self.schema = dict(schema)
        else:
            raise ValueError("Invalid Schema")

        if table_name in self.schema:
            table_schema = self.schema.get(table_name)
            assert isinstance(table_schema, dict)
            self.table_schema = table_schema
        else:
            self.table_schema = {}

        dependencies = extract_dependencies(table_name, year)
        dependencies = [
            table for table, props in dependencies.items() if "size" in props
        ]
        self.table_handler = TableHandler(dependencies, year, settings)

    def load(self, table_name: str | None = None) -> pd.DataFrame:
        """Load the table.

        Builds the table according to the configured settings.
        Will attempt to read from cache first before building
        dynamically from a schema.

        Parameters
        ----------
        table_name : str, optional
            Table to load, will use instance table_name if not specified

        Returns
        -------
        table : DataFrame
            The loaded table data

        """
        table_name = self.table_name if table_name is None else table_name

        if table_name in original_tables:
            table = self.table_handler[table_name]
            if not table.empty and (table_name in self.schema):
                table = self._apply_schema(table, table_name)
        elif self.schema[table_name].get("cache_result", False):
            try:
                table = self.read_cached_table(table_name)
            except FileNotFoundError:
                table = self._construct_schema_based_table(table_name)
                self.save_cache(table, table_name)
        else:
            table = self._construct_schema_based_table(table_name)
        return table

    def read_cached_table(
        self,
        table_name: str,
    ) -> pd.DataFrame:
        """Read a cached table if dependencies are unchanged.

        Checks that the dependencies of the cached table match the
        current dependencies before reading from the cached parquet file.

        Raises FileNotFoundError if dependencies have changed.

        Parameters
        ----------
        table_name : str
            Name of table to read from cache

        Returns
        -------
        table : DataFrame
            The cached table data

        Raises
        ------
        FileNotFoundError
            If cached dependencies are out of date

        """
        if not self.check_table_dependencies(table_name):
            raise FileNotFoundError
        file_name = f"{table_name}_{self.year}.parquet"
        file_path = defaults.cached_data.joinpath(file_name)
        table = pd.read_parquet(file_path)
        return table

    def check_table_dependencies(self, table_name: str) -> bool:
        """Check if cached dependencies match current dependencies.

        Compares the dependencies recorded in the cache metadata file
        to the currently extracted dependencies for the table.

        Parameters
        ----------
        table_name : str
            Table name to check dependencies for

        Returns
        -------
        match : bool
            True if dependencies match, False otherwise

        """
        file_name = f"{table_name}_{self.year}_metadata.yaml"
        cach_metadata_path = defaults.cached_data.joinpath(file_name)
        with open(cach_metadata_path, encoding="utf-8") as file:
            cach_metadata = yaml.safe_load(file)
        file_dependencies = cach_metadata["dependencies"]
        current_dependencies = extract_dependencies(table_name, self.year)
        return file_dependencies == current_dependencies

    def save_cache(
        self,
        table: pd.DataFrame,
        table_name: str,
    ) -> None:
        """Save table to cache along with metadata.

        Saves the table to a parquet file and saves metadata about
        dependencies to a yaml file.

        Parameters
        ----------
        table : DataFrame
            Table data to cache
        table_name : str
            Name of table being cached

        """
        defaults.cached_data.mkdir(parents=True, exist_ok=True)
        file_name = f"{table_name}_{self.year}.parquet"
        file_path = defaults.cached_data.joinpath(file_name)
        file_name = f"{table_name}_{self.year}_metadata.yaml"
        cach_metadata_path = defaults.cached_data.joinpath(file_name)
        file_metadata = {"dependencies": extract_dependencies(table_name, self.year)}
        with open(cach_metadata_path, mode="w", encoding="utf-8") as file:
            yaml.safe_dump(file_metadata, file)
        table.to_parquet(file_path, index=False)

    def _apply_schema(
        self,
        table: pd.DataFrame,
        table_name: str,
    ):
        if "instructions" not in self.schema[table_name]:
            return table

        steps = self.schema[table_name]["instructions"]
        assert isinstance(steps, list)
        table = Pipeline(
            table=table, steps=steps, table_name=table_name, year=self.year
        ).run()
        return table

    def _construct_schema_based_table(self, table_name: str) -> pd.DataFrame:
        if table_name not in self.schema:
            raise KeyError(f"Table name {table_name} is not available in schema")
        table_names = self.schema[table_name]["table_list"]
        assert isinstance(table_names, (str, list))

        table_list = self._collect_schema_tables(table_names)

        table = pd.concat(table_list)
        table = self._apply_schema(table, table_name)
        return table

    def _collect_schema_tables(
        self, table_names: str | list[str]
    ) -> list[pd.DataFrame]:
        table_names = [table_names] if isinstance(table_names, str) else table_names
        table_list = [self.load(name) for name in table_names]
        table_list = [table for table in table_list if not table.empty]
        return table_list


def create_table(
    table_name: str,
    years: _Years,
    settings: LoadTableSettings | None = None,
) -> pd.DataFrame:
    """Construct a table by loading it for multiple years.

    Loads the specified table for each year in the provided
    range of years. Concatenates the individual tables into
    one table indexed by year.

    Parameters
    ----------
    table_name : str
        Name of table to load
    years : int or list of int
        Years to load table for
    settings : LoadTable, optional
        Settings for how to load each table

    Returns
    -------
    table : DataFrame
        Table concatenated across specified years

    """
    table_list = []
    for year in utils.parse_years(years):
        table = TableFactory(table_name, year, settings).load()
        table_list.append(table)
    table = pd.concat(table_list)
    return table


def load_weights(
    year: int,
    adjust_for_household_size: bool = False,
    method: Literal["default", "external", "household_info"] = "default",
) -> pd.Series:
    """Load sample weights for a given year.

    Loads weights from different sources based on the year and specified
    method. Weights can be multiplied by the number of household members
    if multiply_members=True.

    Parameters
    ----------
    year : int
        Year to load weights for
    multiply_members : bool, default False
        Whether to multiply weights by the number of household members
    method : {"default", "external", "household_info"}, default "default"
        Where to load the weights from:

        "default": Use "external" for years <= 1395, "household_info"
        for later years

        "external": Load from external parquet file

        "household_info": Load from household_information table

    Returns
    -------
    weights : pd.Series
        Sample weights indexed by household ID

    """
    if method == "default":
        if year <= 1395:
            method = "external"
        else:
            method = "household_info"
    elif method in ["external", "household_info"]:
        pass
    else:
        raise ValueError("Method is not valid")

    if method == "external":
        weights = _load_from_external_data(year)
    else:
        weights = _load_from_household_info(year)

    if adjust_for_household_size:
        members = (
            create_table("Number_of_Members", years=year)
            .set_index("ID")
            .loc[:, "Members"]
        )
        weights, members = weights.align(members, join="left")
        weights = weights.mul(members)

    weights = weights.rename(defaults.columns.weight)
    return weights


def _load_from_household_info(year) -> pd.Series:
    loader = TableFactory("household_information", year)
    hh_info = loader.load()
    weights = hh_info.set_index("ID")["Weight"]
    return weights


def _load_from_external_data(year) -> pd.Series:
    weights = external_data.load_table("weights", reset_index=False)
    weights = weights.loc[(year), "Weight"]
    assert isinstance(weights, pd.Series)
    return weights


def add_weights(
    table: pd.DataFrame,
    adjust_for_household_size: bool = False,
    year_column_name: str = "Year",
) -> pd.DataFrame:
    """Add sample weights to a table of data.

    Parameters
    ----------
    table : pd.DataFrame
        Input data table, containing a column of year values
    adjust_for_household_size : bool, default False
        Whether to adjust weights by household size
    year_column_name : str, default "Year"
        Name of column in `table` that contains the year

    Returns
    -------
    table : pd.DataFrame
        Input `table` with 'Weight' column added

    """
    years = decoder.extract_column(table, year_column_name)
    years = years.drop_duplicates()
    weights_list = []
    for year in years:
        weights_list.append(load_weights(year, adjust_for_household_size))
    weights = pd.concat(weights_list, axis="index", keys=years, names=["Year", "ID"])
    table = table.join(weights, on=["Year", "ID"])
    return table



"""
Metadata module
"""
import functools
import re

from pathlib import Path
from typing import Any, Literal, Callable, Iterable, get_args

from pydantic import BaseModel
import yaml


PACKAGE_DIRECTORY = Path(__file__).parents[1]
ROOT_DIRECTORT = Path().absolute()


_Attribute = Literal["Urban_Rural", "Province", "County"]
_Years = int | Iterable[int] | str | Literal["all", "last"]

_GeneralTable = Literal[
    "household_information",
    "census_month",
    "members_properties",
    "house_specifications",
]

_ExpenditureTable = Literal[
    "food",
    "tobacco",
    "cloth",
    "home",
    "furniture",
    "medical",
    "transportation",
    "communication",
    "entertainment",
    "education",
    "hotel",
    "miscellaneous",
    "durable",
    "investment",
]

_IncomeTable = Literal[
    "employment_income",
    "self_employed_income",
    "other_income",
    "subsidy",
    "public_employment_income",
    "private_employment_income",
    "agricultural_self_employed_income",
    "non_agricultural_self_employed_income",
    "old_other_income",
]

_OriginalTable = Literal[_GeneralTable, _ExpenditureTable, _IncomeTable]

_StandardTable = Literal[
    "Weights",
    "Number_of_Members",
    "Equivalence_Scale",
    "Original_Expenditures",
    "Expenditures",
    "Total_Expenditure",
    "Original_Outlays",
    "Outlays",
    "Total_Outlay",
    "Imputed_Rent",
    "Incomes",
    "Income_Breakdown",
    "Members_Income_Breakdown",
    "Total_Income",
    "Members_Total_Income",
]

_Table = Literal[_OriginalTable, _StandardTable]


general_tables: tuple[_GeneralTable] = get_args(_GeneralTable)  # type: ignore
expenditure_tables: tuple[_ExpenditureTable] = get_args(_ExpenditureTable)  # type: ignore
original_tables: tuple[_OriginalTable] = get_args(_OriginalTable)  # type: ignore
standard_tables: tuple[_StandardTable] = get_args(_StandardTable)  # type: ignore


_Province = Literal[
    "Markazi",
    "Gilan",
    "Mazandaran",
    "East_Azerbaijan",
    "West_Azerbaijan",
    "Kermanshah",
    "Khuzestan",
    "Fars",
    "Kerman",
    "Razavi_Khorasan",
    "Isfahan",
    "Sistan_and_Baluchestan",
    "Kurdistan",
    "Hamadan",
    "Chaharmahal_and_Bakhtiari",
    "Lorestan",
    "Ilam",
    "Kohgiluyeh_and_Boyer-Ahmad",
    "Bushehr",
    "Zanjan",
    "Semnan",
    "Yazd",
    "Hormozgan",
    "Tehran",
    "Ardabil",
    "Qom",
    "Qazvin",
    "Golestan",
    "North_Khorasan",
    "South_Khorasan",
    "Alborz",
]


_Groupby = Literal["Urban_Rural", "Province", "County", "Decile", "Percentile"]
_EquivalenceScale = Literal[
    "Household", "Per_Capita", "OECD", "OECD_Modified", "Square_Root"
]


def open_yaml(
    path: Path | str,
    location: Literal["package", "root"] = "package",
    interpreter: Callable[[str], str] | None = None,
):
    """
    Read the contents of a YAML file relative to the root directory and return it as a dictionary.

    :param path: The path to the YAML file, relative to the root directory.
    :type path: str

    :return: The contents of the YAML file as a dictionary.
    :rtype: dict

    :raises yaml.YAMLError: If there is an error parsing the YAML file.

    """
    path = Path(path) if isinstance(path, str) else path
    if path.is_absolute():
        pass
    elif location == "root":
        path = ROOT_DIRECTORT.joinpath(path)
    else:
        path = PACKAGE_DIRECTORY.joinpath(path)

    with open(path, mode="r", encoding="utf8") as yaml_file:
        yaml_text = yaml_file.read()
    if interpreter is not None:
        yaml_text = interpreter(yaml_text)
    yaml_content = yaml.safe_load(yaml_text)
    return yaml_content


def flatten_dict(dictionary: dict) -> dict[tuple[Any, ...], Any]:
    flattened_dict = {}
    for key, value in dictionary.items():
        if isinstance(value, dict):
            flattend_value = flatten_dict(value)
            for sub_key, sub_value in flattend_value.items():
                flattened_dict[(key,) + sub_key] = sub_value
        else:
            flattened_dict[(key,)] = value
    return flattened_dict


def collect_settings() -> dict[tuple[Any, ...], Any]:
    sample_settings_path = PACKAGE_DIRECTORY.joinpath("config", "default_settings.yaml")
    _settings = flatten_dict(open_yaml(sample_settings_path))

    package_settings_path = PACKAGE_DIRECTORY.joinpath(_settings[("package_settings",)])
    if package_settings_path.exists():
        package_settings = flatten_dict(open_yaml(package_settings_path))
        _update_settings(_settings, package_settings)

    root_setting_path = ROOT_DIRECTORT.joinpath(_settings[("local_settings",)])
    if root_setting_path.exists():
        root_settings = flatten_dict(open_yaml(root_setting_path))
        _update_settings(_settings, root_settings)

    return _settings


def _update_settings(_settings, new_settings):
    for key, value in new_settings.items():
        if key in _settings:
            _settings[key] = value


settings = collect_settings()


class DefaultColumns(BaseModel):
    year: str = settings[("columns", "year")]
    household_id: str = settings[("columns", "household_id")]
    commodity_code: str = settings[("columns", "commodity_code")]
    job_code: str = settings[("columns", "job_code")]
    weight: str = settings[("columns", "weight")]

    nominals: list = settings[("nominal_columns",)]
    groupby: list[_Groupby] = settings[("groupby_columns",)]


class Defaults(BaseModel):
    # online directory
    online_dir: str = settings[("online_directory",)]

    # local directory
    package_dir: Path = PACKAGE_DIRECTORY
    root_dir: Path = ROOT_DIRECTORT

    if Path(settings[("local_directory",)]).is_absolute():
        local_dir: Path = Path(settings[("local_directory",)])
    elif settings[("in_root",)]:
        local_dir: Path = root_dir.joinpath(settings[("local_directory",)])
    else:
        local_dir: Path = package_dir.joinpath(settings[("local_directory",)])  # type: ignore

    archive_files: Path = local_dir.joinpath(settings[("archive_files",)])
    unpacked_data: Path = local_dir.joinpath(settings[("unpacked_data",)])
    extracted_data: Path = local_dir.joinpath(settings[("extracted_data",)])
    processed_data: Path = local_dir.joinpath(settings[("processed_data",)])
    external_data: Path = local_dir.joinpath(settings[("external_data",)])
    maps: Path = local_dir.joinpath(settings[("maps",)])
    cached_data: Path = local_dir.joinpath(settings[("cached_data",)])

    first_year: int = settings[("first_year",)]
    last_year: int = settings[("last_year",)]

    columns: DefaultColumns = DefaultColumns()

    def model_post_init(self, __contex=None) -> None:
        self.archive_files.mkdir(parents=True, exist_ok=True)
        self.unpacked_data.mkdir(parents=True, exist_ok=True)
        self.extracted_data.mkdir(parents=True, exist_ok=True)
        self.processed_data.mkdir(parents=True, exist_ok=True)
        self.external_data.mkdir(parents=True, exist_ok=True)
        self.maps.mkdir(parents=True, exist_ok=True)
        self.cached_data.mkdir(parents=True, exist_ok=True)


class Metadata:
    """
    A dataclass for accessing metadata used in other parts of the project.

    """

    metadata_files = [
        "instruction",
        "tables",
        "maps",
        "household",
        "commodities",
        "occupations",
        "schema",
        "other",
        "external_data",
    ]
    instruction: dict[str, Any]
    tables: dict[str, Any]
    maps: dict[str, Any]
    household: dict[str, Any]
    commodities: dict[str, Any]
    occupations: dict[str, Any]
    schema: dict[str, Any]
    other: dict[str, Any]
    external_data: dict[str, Any]

    def __init__(self) -> None:
        self.reload()

    def reload(self):
        for file_name in self.metadata_files:
            self.reload_file(file_name)

    def reload_file(self, file_name):
        package_metadata_path = settings[("package_metadata", file_name)]
        local_metadata_path = ROOT_DIRECTORT.joinpath(
            settings[("local_metadata", file_name)]
        )
        interpreter = self.get_interpreter(file_name)
        _metadata: dict = open_yaml(package_metadata_path, interpreter=interpreter)
        interpreter = self.get_interpreter(file_name, _metadata)
        if local_metadata_path.exists():
            local_metadata = open_yaml(local_metadata_path, interpreter=interpreter)
            _metadata.update(local_metadata)
        setattr(self, file_name, _metadata)

    def get_interpreter(
        self, file_name: str, context: dict | None = None
    ) -> Callable[[str], str] | None:
        context = context or {}
        if f"{file_name}_interpreter" in dir(self):
            interpreter = getattr(self, f"{file_name}_interpreter")
            interpreter = functools.partial(interpreter, context=context)
        else:
            interpreter = None
        return interpreter

    @staticmethod
    def commodities_interpreter(yaml_text: str, context: dict) -> str:
        context.update(yaml.safe_load(re.sub("{{.*}}", "", yaml_text)))
        placeholders_list: list[str] = re.findall(r"{{\s*(.*)\s*}}", yaml_text)
        mapping = {}
        for placeholder in placeholders_list:
            parts = placeholder.split(".")
            if len(parts) == 1:
                mapping[placeholder] = context[parts[0]]["items"]
            elif len(parts) == 2:
                mapping[placeholder] = context[parts[0]]["items"][parts[1]]
            else:
                raise ValueError
        for placeholder, value in mapping.items():
            yaml_text = yaml_text.replace("{{" + placeholder + "}}", str(value))
        return yaml_text


metadata = Metadata()
defaults = Defaults()



from .quantile import (
    calculate_quantile,
    add_quantile,
    add_percentile,
    add_decile,
)
from .average import weighted_average, average_table


__all__ = [
    "calculate_quantile",
    "add_quantile",
    "add_percentile",
    "add_decile",
    "weighted_average",
    "average_table",
]



"""
DataFrame extention
"""
import pandas as pd

from .core import decoder


@pd.api.extensions.register_dataframe_accessor("view")
class ViewAccessor:
    def __init__(self, pandas_obj: pd.DataFrame):
        self._validate(pandas_obj)
        self._obj = pandas_obj
        self._views = None

    @staticmethod
    def _validate(obj):
        pass

    @property
    def views(self) -> list[str] | None:
        return self._views

    @views.setter
    def views(self, value: list[str]):
        self._views = value

    def __getitem__(self, value: str) -> pd.DataFrame:
        if self._views is not None and value in self._views:
            settings = decoder.DecoderSettings(
                name=value, classification_type="commodity"
            )
            return decoder.Decoder(self._obj, settings).add_classification()
        raise KeyError(f"{value} is not a view of the current table")



"""HBSIR library utility functions"""
from typing import Any

from .seven_zip_utils import extract as sevenzip
from .download_utils import download, download_map, download_processed_data
from .parsing_utils import parse_years, construct_table_year_pairs
from .metadata_utils import (
    MetadataVersionResolver,
    MetadataCategoryResolver,
    MetadataResolverSettings,
)
from .argham import Argham


__all__ = [
    "sevenzip",
    "download",
    "download_map",
    "download_processed_data",
    "parse_years",
    "construct_table_year_pairs",
    "Argham",
    "resolve_metadata",
]


_Default = Any


# pylint: disable=too-many-arguments
# pylint: disable=unused-argument
def resolve_metadata(
    versioned_metadata: dict,
    year: int,
    categorize: bool = False,
    year_range: tuple[int, int] = _Default,
    year_keyword: str = _Default,
    version_keyword: str = _Default,
    items_keyword: str = _Default,
    category_keyword: str = _Default,
    item_key_name: str = _Default,
) -> ...:
    """Resolves metadata for the given year.

    Resolves the versioned metadata to the specified year. Optionally
    categorizes the resolved metadata if has_category is True.

    Allows overriding default settings via keyword arguments.

    Parameters
    ----------
    versioned_metadata : dict
        Raw metadata dictionary with embedded version info.

    year : int
        Year to resolve metadata for.

    categorize : bool, optional
        Whether to categorize resolved metadata.
        Default is False.

    Additional keywords:
        Used to override default versioning settings.

    Returns
    -------
    dict
        Resolved metadata dictionary for the specified year.
        Categorized if categorize is True.
    """
    setting_parameters = {
        key: value
        for key, value in locals().items()
        if (value is not _Default)
        and (value not in ["versioned_metadata", "year", "has_category"])
    }
    settings = MetadataResolverSettings(**setting_parameters)

    if categorize:
        resolver = MetadataCategoryResolver(versioned_metadata, year, settings)
        resolved_metadata = resolver.categorize_metadata()
    else:
        resolver = MetadataVersionResolver(versioned_metadata, year, settings)
        resolved_metadata = resolver.get_version()

    if isinstance(resolved_metadata, dict):
        resolved_metadata.update({"year": year})

    return resolved_metadata



"""For maintainer"""
import concurrent.futures
from pathlib import Path
import tomllib
from typing import Iterable, Literal

import requests
from tqdm import tqdm
import boto3

from .api import setup
from .core.metadata_reader import (
    defaults,
    _OriginalTable,
    _Years,
)
from . import utils


def update_online_files(recreate=False, replace=False) -> None:
    if recreate:
        setup(years="all", method="create", table_names="all", replace=replace)
    for directories in [
        (defaults.processed_data, "parquet_files"),
        (defaults.external_data, "external_data"),
    ]:
        update_online_directory(*directories)


def update_online_directory(local_directory: Path, online_directory: str) -> None:
    file_stats = _scan_directory(local_directory, online_directory)
    to_up_files = [file for file, stat in file_stats.items() if stat != "UpToDate"]
    with concurrent.futures.ThreadPoolExecutor(max_workers=6) as executor:
        for file in to_up_files:
            executor.submit(
                _publish_parquet_file, local_directory.joinpath(file), online_directory
            )


def _scan_directory(local_directory: Path, online_directory: str) -> dict:
    futures_dict = {}
    with concurrent.futures.ThreadPoolExecutor(max_workers=6) as executor:
        for path in local_directory.iterdir():
            if path.is_dir():
                continue
            futures_dict.update(
                {executor.submit(_get_file_stats, path, online_directory): path}
            )
    file_stats = {
        futures_dict[future].name: future.result()
        for future in concurrent.futures.as_completed(futures_dict)
    }
    return file_stats


def _get_file_stats(
    path: Path, online_directory: str
) -> Literal["Missing", "UpToDate", "OutDated"]:
    online_file_size = _get_file_size_online_directory(path.name, online_directory)
    local_size = path.stat().st_size

    if online_file_size is None:
        return "Missing"
    if abs(online_file_size - local_size) < 5:
        return "UpToDate"
    return "OutDated"


def _get_file_size_online_directory(file_name: str, directory):
    url = f"{defaults.online_dir}/{directory}/{file_name}"
    response = requests.head(url, timeout=10)
    try:
        return int(response.headers["Content-Length"])
    except KeyError:
        return None


def publish_processed_table(
    table_name: _OriginalTable | Iterable[_OriginalTable] | None = None,
    years: _Years = "last",
    online_directory: str = "parquet_files",
) -> None:
    assert isinstance(table_name, str)
    table_year = utils.construct_table_year_pairs(table_name, years)
    pbar = tqdm(total=len(table_year), desc="Preparing ...", unit="Table")
    for _table_name, year in table_year:
        pbar.update()
        pbar.desc = f"Uploading table: {_table_name}, for year: {year}"
        file_name = f"{year}_{_table_name}.parquet"
        file_path = defaults.processed_data.joinpath(file_name)
        _publish_parquet_file(file_path, online_directory)
    pbar.close()


def _publish_parquet_file(file_path: Path, directory: str) -> None:
    url = f"HBSIR/{directory}/{file_path.name}"
    _upload_file_to_online_directory(file_path, url)


def _upload_file_to_online_directory(file_path, file_name):
    bucket = _get_bucket()

    with open(file_path, "rb") as file:
        bucket.put_object(ACL="public-read", Body=file, Key=file_name)


def _get_bucket(bucket_name="sdac"):
    with open("tokens.toml", "rb") as file:
        token = tomllib.load(file)["arvan"]
    s3_resource = boto3.resource(
        "s3",
        endpoint_url="https://s3.ir-tbz-sh1.arvanstorage.ir",
        aws_access_key_id=token["access_key"],
        aws_secret_access_key=token["secret_key"],
    )
    bucket = s3_resource.Bucket(bucket_name)  # type: ignore
    return bucket
